@@ -24,13 +24,17 @@
 #' @param extra_packages Additional Python packages to install along with
 #'   TensorFlow.
 #'
+#' @param restart_session Restart R session after installing (note this will
+#'   only occur within RStudio).
+#'
 #' @importFrom jsonlite fromJSON
 #'
 #' @export
 install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system"),
                                conda = "auto",
                                version = "default",
-                               extra_packages = NULL) {
+                               extra_packages = NULL,
+                               restart_session = TRUE) {
 
   # verify os
   if (!is_windows() && !is_osx() && !is_linux()) {
@@ -59,7 +63,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
   ver <- parse_tensorflow_version(version)
   version <- ver$version
   gpu <- ver$gpu
-  package_url <- ver$package_url
+  packages <- ver$packages
 
   # flags indicating what methods are available
   method_available <- function(name) method %in% c("auto", name)
@@ -82,7 +86,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
         stop("Conda installation failed (no conda binary found)\n", call. = FALSE)
 
       # do install
-      install_tensorflow_conda(conda, version, gpu, package_url, extra_packages)
+      install_tensorflow_conda(conda, version, gpu, packages, extra_packages)
 
     } else {
 
@@ -100,7 +104,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
       # if we don't have pip and virtualenv then try for conda if it's allowed
       if ((!have_pip || !have_virtualenv) && have_conda) {
 
-        install_tensorflow_conda(conda, version, gpu, package_url, extra_packages)
+        install_tensorflow_conda(conda, version, gpu, packages, extra_packages)
 
 
       # otherwise this is either an "auto" installation w/o working conda
@@ -153,7 +157,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
         }
 
         # do the install
-        install_tensorflow_virtualenv(python, virtualenv, version, gpu, package_url, extra_packages)
+        install_tensorflow_virtualenv(python, virtualenv, version, gpu, packages, extra_packages)
 
       }
     }
@@ -197,7 +201,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
       }
 
       # do the install
-      install_tensorflow_conda(conda, version, gpu, package_url, extra_packages)
+      install_tensorflow_conda(conda, version, gpu, packages, extra_packages)
 
     } else if (identical(method, "system")) {
 
@@ -212,7 +216,7 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
       # do system installation
       python <- python_system_version$executable_path
       pip <- file.path(python_system_version$install_path, "Scripts", "pip.exe")
-      install_tensorflow_windows_system(python, pip, version, gpu, package_url, extra_packages)
+      install_tensorflow_windows_system(python, pip, version, gpu, packages, extra_packages)
 
     } else {
       stop("Invalid/unexpected installation method '", method, "'",
@@ -220,11 +224,15 @@ install_tensorflow <- function(method = c("auto", "virtualenv", "conda", "system
     }
   }
 
-  cat("\nInstallation of TensorFlow complete.\n\n")
+  cat("\nInstallation complete.\n\n")
+
+  if (restart_session && rstudioapi::hasFun("restartSession"))
+    rstudioapi::restartSession()
+
   invisible(NULL)
 }
 
-install_tensorflow_conda <- function(conda, version, gpu, package_url, extra_packages = NULL) {
+install_tensorflow_conda <- function(conda, version, gpu, packages, extra_packages = NULL) {
 
   # create conda environment if we need to
   envname <- "r-tensorflow"
@@ -236,8 +244,38 @@ install_tensorflow_conda <- function(conda, version, gpu, package_url, extra_pac
   }
   else {
     cat("Creating", envname, "conda environment for TensorFlow installation...\n")
-    packages <- ifelse(is_windows(), "python=3.6", "python")
-    python <- conda_create(envname, packages = packages, conda = conda)
+    python_packages <- ifelse(is_windows(), "python=3.6", "python")
+    python <- conda_create(envname, packages = python_packages, conda = conda)
+  }
+
+  # Short circuit to install everything with conda when no custom packages
+  # (typically tf-nightly or a daily build URL) and no gpu-enabled build
+  # is requested (conda doesn't currently have GPU enabled builds.
+  #
+  # This avoids any use of pip, which addresses the following issue:
+  # https://github.com/rstudio/keras/issues/147
+  #
+  # This issue is in turn created by two other issues:
+  #
+  # 1) TensorBoard appears to rely on an older version of html5lib which is
+  #    force installed, and which as a result breaks pip:
+  #    https://github.com/tensorflow/tensorboard/issues/588
+  #
+  # 2) Anaconda 5.0.0 is unable to recover from this because the installation
+  #    of the old version of html5lib actually propagates to the root
+  #    environment, which permantely breaks pip for *all* conda environments:
+  #    https://github.com/conda/conda/issues/6079
+  #
+  # Hopefully these two issues will be addressed and we can return to using
+  # pip in all scenarios (as that is the officially supported version)
+  #
+  if (is_windows() && is.null(packages) && gpu == FALSE) {
+    conda_forge_install(
+      envname,
+      tf_pkgs(version, gpu, packages, extra_packages = extra_packages),
+      conda = conda
+    )
+    return(invisible(NULL))
   }
 
   # determine tf version
@@ -271,10 +309,10 @@ install_tensorflow_conda <- function(conda, version, gpu, package_url, extra_pac
   # determine arch
   arch <- ifelse(is_windows(), "win_amd64", ifelse(is_osx(), "any", "linux_x86_64"))
 
-  # determine package_url
-  if (is.null(package_url)) {
+  # determine packages url if necessary
+  if (is.null(packages)) {
     platform <- ifelse(is_windows(), "windows", ifelse(is_osx(), "mac", "linux"))
-    package_url <- sprintf(
+    packages <- sprintf(
       "https://storage.googleapis.com/tensorflow/%s/%s/tensorflow%s-%s-%s-%s.whl",
       platform,
       ifelse(gpu, "gpu", "cpu"),
@@ -287,7 +325,7 @@ install_tensorflow_conda <- function(conda, version, gpu, package_url, extra_pac
 
   # install base tensorflow using pip
   cat("Installing TensorFlow...\n")
-  conda_install(envname, package_url, pip = TRUE, conda = conda)
+  conda_install(envname, packages, pip = TRUE, conda = conda)
 
   # install additional packages
   conda_install(envname, tf_extra_pkgs(), conda = conda)
@@ -306,7 +344,27 @@ install_tensorflow_conda <- function(conda, version, gpu, package_url, extra_pac
   }
 }
 
-install_tensorflow_virtualenv <- function(python, virtualenv, version, gpu, package_url, extra_packages = NULL) {
+
+conda_forge_install <- function(envname, packages, conda = "auto") {
+
+  # resolve conda binary
+  conda <- conda_binary(conda)
+
+  # use native conda package manager with conda forge enabled
+  result <- system2(conda, shQuote(c("install", "-c", "conda-forge", "--yes", "--name", envname, packages)))
+
+  # check for errors
+  if (result != 0L) {
+    stop("Error ", result, " occurred installing packages into conda environment ",
+         envname, call. = FALSE)
+  }
+
+  invisible(NULL)
+}
+
+
+
+install_tensorflow_virtualenv <- function(python, virtualenv, version, gpu, packages, extra_packages = NULL) {
 
   # determine python version to use
   is_python3 <- python_version(python) >= "3.0"
@@ -360,12 +418,12 @@ install_tensorflow_virtualenv <- function(python, virtualenv, version, gpu, pack
   pip_install("setuptools", "Upgrading setuptools")
 
   # install tensorflow and related dependencies
-  pkgs <- tf_pkgs(version, gpu, package_url, scipy = TRUE, extra_packages = extra_packages)
+  pkgs <- tf_pkgs(version, gpu, packages, scipy = TRUE, extra_packages = extra_packages)
   pip_install(pkgs, "Installing TensorFlow")
 }
 
 
-install_tensorflow_windows_system <- function(python, pip, version, gpu, package_url, extra_packages = NULL) {
+install_tensorflow_windows_system <- function(python, pip, version, gpu, packages, extra_packages = NULL) {
 
   # ensure pip is up to date
   cat("Preparing for installation (updating pip if necessary)\n")
@@ -376,7 +434,7 @@ install_tensorflow_windows_system <- function(python, pip, version, gpu, package
   # install tensorflow and dependencies (don't install scipy b/c it requires
   # native code compilation)
   cat("Installing TensorFlow...\n")
-  pkgs <- tf_pkgs(version, gpu, package_url, scipy = FALSE, extra_packages = extra_packages)
+  pkgs <- tf_pkgs(version, gpu, packages, scipy = FALSE, extra_packages = extra_packages)
   result <- system2(pip, c("install", "--upgrade --ignore-installed",
                            paste(shQuote(pkgs), collapse = " ")))
   if (result != 0L)
@@ -392,7 +450,7 @@ parse_tensorflow_version <- function(version) {
   ver <- list(
     version = "latest",
     gpu = FALSE,
-    package_url = NULL
+    packages = NULL
   )
 
   # full url provided
@@ -413,9 +471,12 @@ parse_tensorflow_version <- function(version) {
     ver$gpu <-TRUE
 
   # full path to installer binary
-  } else if (grepl("^http", version)) {
+  } else if (grepl("^.*\\.whl$", version)) {
 
-    ver$package_url <- version
+    if (grepl("^http", version))
+      ver$packages <- version
+    else
+      ver$packages <- normalizePath(version)
 
   # another version
   } else {
@@ -424,6 +485,14 @@ parse_tensorflow_version <- function(version) {
 
   }
 
+  # if it's the nightly version then set packages to nightly[-gpu]
+  # (also add tensorboard since it's not included in tf-nightly)
+  version <- sub("^tf-nightly$", "nightly", version)
+  if (identical(version, "nightly")) {
+    ver$packages <- c(sprintf("tf-nightly%s", ifelse(ver$gpu, "-gpu", "")),
+                      "tensorflow-tensorboard")
+  }
+
   # return
   ver
 }
@@ -459,13 +528,12 @@ python_version <- function(python) {
 
 
 # form list of tf pkgs
-tf_pkgs <- function(version, gpu, package_url, scipy = TRUE, extra_packages = NULL) {
-  package <- package_url
-  if (is.null(package))
-    package <- sprintf("tensorflow%s%s",
+tf_pkgs <- function(version, gpu, packages, scipy = TRUE, extra_packages = NULL) {
+  if (is.null(packages))
+    packages <- sprintf("tensorflow%s%s",
                        ifelse(gpu, "-gpu", ""),
                        ifelse(version == "latest", "", paste0("==", version)))
-  c(package, tf_extra_pkgs(scipy = scipy, extra_packages = extra_packages))
+  c(packages, tf_extra_pkgs(scipy = scipy, extra_packages = extra_packages))
 }
 
 # additional dependencies to install (required by some features of keras)
@@ -545,8 +613,12 @@ install_tensorflow_extras <- function(packages, conda = "auto") {
   }
 
   # perform the installation
-  if (identical(type, "conda"))
-    conda_install("r-tensorflow", packages, pip = TRUE, conda = conda)
+  if (identical(type, "conda")) {
+    if (is_windows())
+      conda_forge_install("r-tensorflow", packages, conda = conda)
+    else
+      conda_install("r-tensorflow", packages, pip = TRUE, conda = conda)
+  }
   else if (identical(type, "virtualenv"))
     virtualenv_install("r-tensorflow", packages)
   else if (identical(type, "system"))
